describe("base.js", function() {
  var gl = (function(){return this;})(); //グローバルオブジェクト

  it("cascading inherit", function() {
      /*カスケード継承 (cascading inherit)に関するスペック*/
      base("$parent").mix( function() {
        this.hoge = function() {
          return 12;
        };
        this.up("$child").mix( function() {
          this.hoge = function() {
            return this.$parent.hoge();
          };
        });
      });
      expect(base("$parent")).toEqual(gl.$parent);
      expect(base("$parent").$child).toEqual(gl.$parent.$child);
      expect(base("$parent").$child.hoge()).toEqual(12);
      expect(base("$parent").hoge()).toEqual(12);
      expect(gl.$parent.$child.hoge()).toEqual(12);
      expect(base("$parent").$child.hoge).not.toEqual(base("$parent").hoge());
  });

  it("a function base", function() {
    var falsyList = ["", null, 0, undefined],
        i = falsyList.length;
    expect(function(){
      base();
    }).toThrow();
    while(i--) {
      expect(function(){
        base(falsyList[i]);
      }).toThrow();
    }
    expect(typeof base("$wo")).toBe("object");
    expect(base("$w")).toBe(base("$w"));
    expect(base("$w")).toBe(gl.$w);
  });

  it("a mix method", function() {
    /*mixメソッドはオブジェクトの合成ができる*/
    expect(typeof base("$a").mix).toBe("function");
    /*メソッドチェーンのチェック*/
    expect(base("$a").mix({})).toBe(base("$a"));
    expect(base("$a").mix(function(){})).toBe(base("$a"));
    expect(base("$a").mix({}).mix({})).toBe(base("$a"));
    expect(base("$a").mix(function(){}).mix({}).mix(function(){})).toBe(base("$a"));

    var falsyList = ["", null, 0, undefined],
        i = falsyList.length;
    expect(function(){
      base("$a").mix();
    }).toThrow();
    while(i--) {
      expect(function(){
        base("$a").mix(falsyList[i]);
      }).toThrow();
    }

    /*argument to object*/
    base("$a").mix({
      a: 12,
      b: {
        d: 12
      },
      c: function(){
        return (this.a+3);
      }
    });
    expect(base("$a").a).toEqual(12);
    expect(base("$a").b.d).toEqual(12);
    expect(base("$a").c()).toEqual(15);

    /*argument to function*/
    base("$b").mix( function() {
      this.a = 12;
      this.b = {
        d: 12
      };
      this.c = function(){
        return (this.a+3);
      };
    });
    expect(base("$b").a).toEqual(12);
    expect(base("$b").b.d).toEqual(12);
    expect(base("$b").c()).toEqual(15);

    base("$a").mix({
      a: 0,
      b: {
        d: -12
      },
      c: function(){
        return (this.a+3);
      }
    });
    expect(base("$a").a).toEqual(0);
    expect(base("$a").b.d).toEqual(-12);
    expect(base("$a").c()).toEqual(3);

    base("$a").mix(Math);
    expect(base("$a").pi).toEqual(Math.pi);

    base("$a").mix({
      a: "hogehoge",
      b: {
        d: 0
      },
      c: function(){
        return (this.a+"hoge");
      }
    });
    expect(base("$a").a).toBe("hogehoge");
    expect(base("$a").b.d).toEqual(0);
    expect(base("$a").c()).toBe("hogehogehoge");

    for (var i=0, arr = [];i<500;++i) {
      arr.push(i);
    }
    for (i=0;i<100;++i) {
      base("$a").mix(arr);
    }
    expect(base("$a")[0]).toEqual(0);
    expect(base("$a")[499]).toEqual(499);
  });

  it("an up method", function() {
    /*upメソッドは、プロトタイプ継承した新しいオブジェクトを作り、引数で指定された名前のプロパティに格納する*/
    expect(typeof base("$c").up).toEqual("function");

    expect(base("$c").up("$a")).not.toBe(base("$a"));

    expect(base("$c").up("$c")).not.toBe(base("$c"));
    /*もう一度繰り返す (reapeat)*/
    expect(base("$c").up("$c")).not.toBe(base("$c"));

    expect(base("$c").$c).toBe(gl.$c.$c);
    expect(base("$c").$c.up("$c")).not.toBe(base("$c"));
    expect(gl.$c.$c.up("$c")).toBe(base("$c").$c.$c.$c);

    /*オブジェクト$cのhogeプロパティの継承*/
    base("$c").hoge = 12;
    expect(base("$c").$c.hoge).toEqual(12);
    expect(base("$c").$c.$c.hoge).toEqual(12);
    expect(base("$c").$c.$c.$c.hoge).toEqual(12);

    /*プロトタイプチェーン上のオブジェクトも合成できるようにする*/
    base("$c").mix({
        aa: 12
      }).up("$1").mix({
        bb: 15
      });
    base("$d").mix(base("$c").$1);
    expect(base("$d").aa).toEqual(12);
    expect(base("$d").bb).toEqual(15);
    expect(base("$d").$1).toBe(base("$c").$1);
    base("$d").mix(function() {
      expect(this.$c.aa).toEqual(12);
      expect(this.$1.aa).toEqual(12);
      expect(this.$1.bb).toEqual(15);
    });

    for (var i=0;i<5000;++i) {
      base("$c").up("$dd");
    }
    expect(base("$c").$dd.hoge).toEqual(12);
    
    /*引数を省略した場合、引数に文字列"$1"を入れたのと同じ効果が得られる*/
    expect(base("$up").$1).toBeUndefined();
    expect(base("$up").up()).toBe(base("$up").$1);
  });

  it("a mix method function", function() {
    /*mixメソッドの引数に指定された関数と、mixメソッドのレシーバオブジェクトについて*/
    base("$mmf").mix(function(_) {
      expect(_).toBe(base("$mmf"));
      expect(this).toBe(_);
    });
  });

  it("an on method", function() {
    /*onメソッドは同一名メソッドの合成ができる*/
    base("$a").on("met", function() {return 12;});
    /*メソッドチェーンかどうかのチェック*/
    expect(base("$a").on("met", function(a) {
          this.a = a;
        }))
        .toBe(base("$a"));
    base("$a").met(12);
    expect(base("$a").a).toEqual(12);

    base("$a").on("met", function(a) {
        this.b = a;
    }).met(12);
    expect(base("$a").a).toEqual(12);
    expect(base("$a").b).toEqual(12);

    /*返り値は、最初に指定しておいたメソッド*/
    expect(base("$a").met()).toEqual(12);

    base("$a").mix( function() {
        this.a = 0;
        this.b = 0;
    });
    base("$a").up("$b").on("met", function() {
          this.a = 15;
          this.b = 15;
        }).met();
    expect(base("$a").$b.a).toEqual(15);
    expect(base("$a").$b.b).toEqual(15);

    base("$a").on("metmet", function() {
        this.a = 12;
        this.b = 12;
    }).metmet();
    expect(base("$a").a).toEqual(12);
    expect(base("$a").b).toEqual(12);
    /*別のmetメソッドを設定しても、影響を与えないかどうかをチェック*/
    base("$a").on("met", function() {
        this.a = 15;
        this.b = 15;
    }).metmet();
    expect(base("$a").a).toEqual(12);
    expect(base("$a").b).toEqual(12);

    expect(base("$a").met()).toEqual(12);
    expect(base("$a").metmet()).toBeUndefined();

    /*境界値のチェック*/
    base("$a").on("o", function(){})
                 .o();
    expect(base("$a").on("o", function(){return 12;})
                 .o()).toBeUndefined();

    /*通常のメソッドが入っていた場合*/
    base("$a").mix( {aa: function() {
          this.__t = 12;
          return 15;
        }})
        .on("aa", function(){
              this.__n = 12;
            });
    expect(base("$a").aa()).toEqual(15);
    expect(base("$a").__t).toEqual(12);
    expect(base("$a").__n).toEqual(12);
    base("$a").mix({
          aaa: function() {
              this.__t = 18;
              this.__n = 18;
              return 15;
          }
        }).up("$b").on("aaa", function() {
            this.__n = 15;
        });
    expect(base("$a").aaa()).toEqual(15);
    expect(base("$a").$b.__n).toEqual(18);
    expect(base("$a").$b.aaa()).toEqual(15);
    expect(base("$a").$b.__n).toEqual(15);

    /*有効同値のチェック*/
    base("$a").mix({
          am: function() {
              this.__t = 18;
              this.__n = 18;
              return 15;
          }
        }).up("$b").up("$c").on("am", function() {
            this.__n = 15;
        });
    expect(base("$a").am()).toEqual(15);
    expect(base("$a").$b.$c.__n).toEqual(18);
    expect(base("$a").$b.$c.am()).toEqual(15);
    expect(base("$a").$b.$c.__n).toEqual(15);
    base("$a").on(
          "am", function(f) {
              this.__t = f;
              return 15;
          }
        ).up("$b").up("$c").on("am", function(f) {
            this.__n = 15;
        });
    expect(base("$a").am(12)).toEqual(15);
    expect(base("$a").__t).toEqual(12);
    expect(base("$a").$b.$c.__n).toEqual(18);
    expect(base("$a").$b.$c.am()).toEqual(15);
    expect(base("$a").$b.$c.__n).toEqual(15);

    /*無効同値のチェック*/
    expect(function() {
        base("$a").on("met");
    }).toThrow();
    expect(function() {
        base("$a").on("met", "met");
    }).toThrow();
    expect(function() {
        base("$a").on("met", 12);
    }).toThrow();
    expect(function() {
        base("$a").on("", function(){});
    }).toThrow();
    
    /*up、on、mix、ofメソッドは自己参照の矛盾を引き出すので、例外を投げておく*/
    var message = "Invalid method name error";
    expect(function() {
        base("$a").on("up", function(){});
    }).toThrow(message);
    expect(function() {
        base("$a").on("on", function(){});
    }).toThrow(message);
    expect(function() {
        base("$a").on("mix", function(){});
    }).toThrow(message);
    expect(function() {
        base("$a").on("of", function(){});
    }).toThrow(message);

    base("$a").on("upon", function(){}); //これは正常に作動

    /*無関係のオブジェクトを返さないようにするためのチェック*/
    this._ddd = {};
    base("_ddd").up("$1");
  });
  
  describe("an 'of' method", function() {
    /*境界条件を調べる*/
    it("limit value analysis", function() {
        expect(typeof base("$fp").of).toBe("function");
        /*メソッドチェーンを作る*/
        expect(base("$fp").of({})).toBe(base("$fp"));

        /*同じオブジェクトで再度、ofメソッドを呼び出すことはできない*/
        expect(function() {
          base("$fp").of({});
        }).toThrow("Reset error");
        /*upメソッドで継承したオブジェクトは再度呼び出せる*/
        base("$fp").up().of( {} );
        
        base("$fp").up().of ( {
          a: 0,
          b: 1,
          c: true,
          d: false,
          e: {},
          f: function() {},
          g: [],
          h: null,
          call: function() {}
          
        } );
        base("$fp").$1.mix( function(_) {
          expect(_.a).toEqual(0);
          expect(_.b).toEqual(1);
          expect(_.c).toBeTruthy();
          expect(_.d).toBeFalsy();
          expect(typeof _.e).toBe("object");
          /*callメソッドのあるオブジェクトや関数が指定された時だけ、未定義値
           * ただし、callメソッドは指定されても未定義とならない*/
          expect(_.f).toBeUndefined();
          expect(typeof _.call).toBe("function");
          expect(_.g.length).toEqual(0);
          expect(_.h).toBeNull();
        } );
    });
    /*有効同値を調べる*/
    it("Equivalence partitioning, the following is the valid partion", function() {
      base("$ffp").of({
        a: 10,
        b: 12,
        c: base("$afp"),
        call: function() {
          return this.a + this.b;
        }
      });
      var _ = base("$ffp");
      expect(_.a).toEqual(10);
      expect(_.b).toEqual(12);
      expect(_.c).toBeUndefined();
      expect(_.call()).toEqual(22);
      /*upメソッドを使えば、再び、ofメソッドを再設定できる*/
       _ = base("$ffp").up();
      expect(_.a).toEqual(10);
      expect(_.b).toEqual(12);
      expect(_.c).toBe(base("$afp"));
      expect(_.call()).toEqual(22);
      _.of({
        a: 11,
        b: 13,
        c: "bfp",
        call: function() {
          return this.a + this.b;
        }
      });
      expect(_.a).toEqual(11);
      expect(_.b).toEqual(13);
      expect(_.c).toBe("bfp");
      expect(_.call()).toEqual(24);
      
      /*mixメソッドでも、オブジェクト継承の後であれば上書きが可能*/
      _.up("$p").mix( {
        a: 12,
        b: 13,
        c: base("$afp")
      } );
      expect(_.$p.a).toEqual(12);
      expect(_.$p.b).toEqual(13);
      expect(_.$p.c).toBe(base("$afp"));
      expect(_.$p.call()).toEqual(25);
    });
    
    /*無効同値を調べる*/
    it("Equivalence partitioning, the following is the invalid partion", function() {
      expect( function() {
          base("$ffp").up("$error").of( {
            call: null
          } )
          .of( {} );
      } ).toThrow("Reset error");
      expect( function() {
          base("$ffp").up("$error").of();
      } ).toThrow("No arguments error");
    } );
  });
  describe("a 'call' method", function() {
    it("limit value analysis", function() {
      expect(typeof base("$bfp").call).toBe("function");
      expect(base("$bfp").call()).toBe(base("$bfp"));
      expect(base("$bfp").up().call()).toBe(base("$bfp").$1);
      base("$bfp").up().of ( {
        a: 0,
        b: 1,
        c: true,
        d: false,
        e: {},
        f: function() {},
        g: [],
        h: null
      } ).call();
      base("$bfp").$1.mix( function(_) {
        expect(_.a).toEqual(0);
        expect(_.b).toEqual(1);
        expect(_.c).toBeTruthy();
        expect(_.d).toBeFalsy();
        expect(typeof _.e).toBe("object");
        /*callメソッドが設定されていないことに注意*/
        expect(_.f).toBeUndefined();
        expect(typeof _.call).toBe("function");
        expect(_.g.length).toEqual(0);
        expect(_.h).toBeNull();
      } );
    });
    it("Equivalence partitioning, the following is the valid partion", function() {
      base("$bfp").of( {
        a: base("$abfp"),
        call: function() {
          return this.a;
        }
      });
      expect(base("$bfp").a).toBeUndefined();
      base("$abfp").of( {
        call: function() {
          return 12;
        }
      } );
      expect(base("$bfp").call()).toEqual(12);
      /*再度呼び出しても同じ結果となる*/
      expect(base("$bfp").call()).toEqual(12);
      base("$bfp").up("$a").of( {
        a: base("$abfp"),
        b: base("$bbfp")
      } );
      expect(base("$bfp").$a.call()).toEqual(12);
      /*今度は束縛変数としてbプロパティの値を返す*/
      base("$bfp").up("$b").of( {
        b: base("$bbfp"),
        call: function() {
          return this.b;
        }
      } );
      base("$bbfp").of( {
        call: function() {
          return 15;
        }
      } );
      expect(base("$bfp").$b.call()).toEqual(15);
      /*再実行*/
      for (var i=0;i<1000;++i) {
        expect(base("$bfp").$b.call()).toEqual(15);
      }
      /*callメソッドを何も設定していない場合*/
      expect(base("$cfp").of({a:base("$bfp")}).call()).toBe(base("$cfp"));
      expect(base("$cfp").up("$bfp").of({a:base("$bfp")}).call()).toBe(base("$cfp").$bfp);
      
      /*循環参照は防ぐようにする*/
      expect(base("$bfp").$b.up().of( {b: base("$bfp").$b.$1} ).call()).toBe(15);
      
      /*ofメソッドで指定されたfuncメソッドなどは、callメソッドで呼び出し時に実行される（thisはcallメソッドのレシーバ)*/
      expect( base("$cfp").up("$a").of( {
        func: function() {
          return this.$a;
        },
        call: function() {
          return this.func;
        }
      } ).call() ).toBe(base("$cfp").$a);
      
      /*遅延評価*/
      base("$bfp").up("$s").of( {
        lazy: {
          a: base("$cfp")
        },
        call: function() {
          return this.lazy.a;
        }
      } );
      expect(base("$bfp").$s.call()).toBe(base("$cfp"));
    });
    
    /*無効同値を調べる*/
    it("Equivalence partitioning, the following is the invalid partion", function() {
      base("$efp").of( {
        a: base("$ee"),
        call: function() {
          return this.a();
        }
      } );
      expect(function() {
        base("$efp").call();
      }).toThrow();
      } );
  });
  
  describe("an 'initToEval' method", function() {
    it("limit value analysis", function() {
      base("$ifp").of(
        {a: function(){return 1;}, b: base("$ifp"),
          call: function() {return 2;}
        }
      ).initToEval();
      expect(base("$ifp").a).toBeUndefined();
      expect(base("$ifp").b).toBeUndefined();
      expect(base("$ifp").b__).toBe(base("$ifp"));
      base("$ifp").call();
      expect(base("$ifp").a).toBe(1);
      expect(base("$ifp").b).toBe(2);
      expect(base("$ifp").b__).toBe(base("$ifp"));
      base("$ifp").initToEval().mix( {a__: function(){return 2;}} );
      expect(base("$ifp").a).toBe(1);
      base("$ifp").call();
      expect(base("$ifp").a).toBe(2);
      expect(base("$ifp").b).toBe(2);
      expect(base("$ifp").b__).toBe(base("$ifp"));
      
    } );
  } );
  
  it("should occur an error, if no arguments", function(){
    var message = "No arguments error";
    expect(function(){
      base();
    }).toThrow(message);
    expect(function(){
      base("$error").mix();
    }).toThrow(message);
    expect(function(){
      base("$error").on();
    }).toThrow(message);
    expect(function(){
      base("$error").of();
    }).toThrow(message);
    message = void 0;
  });
});
